;; Generated package description from ivy-posframe.el  -*- no-byte-compile: t -*-
(define-package "ivy-posframe" "0.5.5" "Using posframe to show Ivy" '((emacs "26.0") (posframe "0.8.0") (ivy "0.13.0")) :keywords '("abbrev" "convenience" "matching" "ivy") :authors '(("Feng Shu" . "tumashu@163.com") ("Naoya Yamashita" . "conao3@gmail.com")) :maintainer '("Feng Shu" . "tumashu@163.com") :url "https://github.com/tumashu/ivy-posframe")
